part of 'application.dart';

/// The ViewModel class is designed to store and manage data in a lifecycle conscious way.
/// ViewModel is a class that is responsible for preparing and managing the data for a Widget tree.
/// It also handles the communication of the Widgets with the rest of the application (e.g. calling the
/// business logic classes).
/// A ViewModel is always created in association with a scope (defined with the use of the related [ViewModelScope]
/// widget) and will be retained as long as the scope is alive (until unmount).
/// The purpose of the ViewModel is to acquire and keep the information that is necessary for your widgets.
/// This widgets should be able to observe changes in the ViewModel.
/// ViewModels usually expose this information via [LiveData].
/// ViewModel's only responsibility is to manage the data for the UI.
/// It should never access your view hierarchy or hold a reference back to the widgets tree.
class ViewModel {
  @protected
  @mustCallSuper
  void onCleared() {}
}

/// Factory used to instantiate your ViewModel.
/// The best practice is defining the ViewModel constructor as private and exposing a static parameterless method that
/// returns the newly constructed instance of your ViewModel. You should pass only that method as [ViewModelFactory]
/// parameter
typedef ViewModelFactory<T extends ViewModel> = T Function(ViewModelProvider provider);

/// The provider interface provides you the ViewModel
abstract class ViewModelProvider {
  const ViewModelProvider._();

  /// Provide a [VM] instance.
  /// The first time you call this method on [VM], [factory] will be called to create the ViewModel instance which
  /// will be retained for future calls until scope disposition
  VM provide<VM extends ViewModel>(ViewModelFactory<VM> factory);

  /// Clears all ViewModel instances that belongs to this provider
  void clear();
}

/// Define a ViewModel scope.
/// All descendants of this widget will have access to the [ViewModelProvider] of this scope through it's context
/// Use cases of scoped viewModels
///     context.viewModel(MyViewModel.factory) // scope: <global>
///
///     ViewModelScope.builder(
///       scope: #my_scope,
///       builder: (context) => context.viewModel(MyViewModel.factory)); // scope: #my_scope
///
///     ViewModelScope.builder(
///       scope: #my_scope,
///       builder: (context) => ViewModelScope.builder(
///         scope: #my_scope_2,
///         builder: (context) => context.viewModel(MyViewModel.factory))); // scope: #my_scope_2
///
///     ViewModelScope.builder(
///       scope: #my_scope,
///       builder: (context) => ViewModelScope.builder(
///         scope: #my_scope_2,
///         builder: (context) => context.viewModel(MyViewModel.factory, scope: #my_scope))); // scope: #my_scope
class ViewModelScope extends InheritedWidget {
  final Symbol scope;

  ViewModelScope({required this.scope, required Widget child}) : super(key: GlobalObjectKey(scope), child: child);

  ViewModelScope.builder({required Symbol scope, required WidgetBuilder builder})
      : this(scope: scope, child: Builder(builder: builder));

  @override
  bool updateShouldNotify(ViewModelScope oldWidget) => oldWidget.scope != scope;

  @override
  InheritedElement createElement() => _ViewModelScopeElement(this);

  @override
  void debugFillProperties(DiagnosticPropertiesBuilder properties) {
    super.debugFillProperties(properties);
    properties.add(DiagnosticsProperty<Symbol>('scope', scope));
  }

  static ViewModelProvider providerOf(BuildContext context, {Symbol? scope}) {
    var element = context.getElementForInheritedWidgetOfExactType<ViewModelScope>() as _ViewModelScopeElement;
    if (scope != null)
      while (element.widget.scope != scope) {
        final parent = element._parent;
        if (parent == null) throw StateError('no scope #${scope.name} associated with current context');
        element = parent;
      }
    return element;
  }
}

class _ViewModelScopeElement extends InheritedElement implements ViewModelProvider {
  final Map<Type, ViewModel> _viewModels = {};

  @override
  ViewModelScope get widget => super.widget as ViewModelScope;

  _ViewModelScopeElement? _parent;

  _ViewModelScopeElement(ViewModelScope widget) : super(widget);

  @override
  void mount(Element? parent, Object? newSlot) {
    super.mount(parent, newSlot);
    _parent = parent?.getElementForInheritedWidgetOfExactType<ViewModelScope>() as _ViewModelScopeElement?;
  }

  @override
  VM provide<VM extends ViewModel>(ViewModelFactory<VM> factory) => (_viewModels[VM] ??= factory(this)) as VM;

  @override
  void unmount() {
    super.unmount();
    clear();
  }

  @override
  void clear() {
    for (final key in _viewModels.keys.toList(growable: false)) _viewModels.remove(key)?.onCleared();
  }
}

extension BuildContextExt on BuildContext {
  VM viewModel<VM extends ViewModel>(ViewModelFactory<VM> factory, {Symbol? scope}) =>
      ViewModelScope.providerOf(this, scope: scope).provide(factory);
}

mixin TaskViewModel on ViewModel {
  final LiveList<Future> activeTasks = LiveList();
  MediatorLiveData<bool>? _loading;

  LiveData<bool> get loading =>
      _loading ??= MediatorLiveData()..addTransformSource<List>(activeTasks, (tasks) => tasks.isNotEmpty);

  @protected
  Future addTask(Future task, RouteArgs args) {
    activeTasks.add(task);
    return task.whenComplete(() => activeTasks.remove(task));
  }

  @override
  void onCleared() {
    super.onCleared();
    activeTasks.clear();
    _loading = null;
  }
}

extension FutureExt on Future {
  void notify(TaskViewModel viewModel, {RouteArgs args = const {}}) => viewModel.addTask(this, args);
}
