import 'package:flutter/foundation.dart';
import 'package:flutter/widgets.dart';
import 'package:flutterx_application/flutterx_application.dart';
import 'package:flutterx_live_data/flutterx_live_data.dart';

/// Interface for localizing your application in a efficient way:
///
/// First: declare your interface:
/// ```dart
/// late Labels labels;
///
/// class Labels extends LabelInterface {
///   static final Set<Labels> supportedLabels = {_it, _en};
///   static final LiveEvent<Labels> onLabel = LiveEvent()..observeForever((value) => labels = value);
///   final String appName;
///
///   const Labels._({
///     required Locale locale,
///     required this.appName,
///   }) : super(locale);
///
/// }
/// ```
///
/// Then define an implementation for every language:
/// ```dart
/// final _en = Labels._(
///   locale: const Locale('en', 'US'),
///   appName: 'Application Demo',
/// );
/// ```
///
/// Finally pass Labels.supportedLabels and Labels.onLabel to corresponding parameters of [runApplication]
///
/// And then you're ready to use labels.appName without worrying about translations
abstract class LabelInterface {
  final Locale locale;

  const LabelInterface(this.locale);
}

/// Used internally by [Application] you may not interact with this class directly
class LabelInterfaceDelegate extends LocalizationsDelegate<LabelInterface> {
  final Set<LabelInterface> labels;
  final ValueSetter<LabelInterface> onLabel;

  const LabelInterfaceDelegate(this.labels, this.onLabel);

  @override
  bool isSupported(Locale locale) => labels.any((label) => label.locale.languageCode == locale.languageCode);

  @override
  Future<LabelInterface> load(Locale locale) {
    final label = labels.firstWhere((label) => label.locale.languageCode == locale.languageCode);
    onLabel(label);
    return SynchronousFuture(label);
  }

  @override
  bool shouldReload(LabelInterfaceDelegate old) => !setEquals(old.labels, labels) || old.onLabel != onLabel;
}

/// Some StatefulWidget may not rebuild automatically, in that case use this mixin
mixin LocaleChangeAware<T extends StatefulWidget> on State<T> {
  late ObserverWrapper<LabelInterface> _observer;

  @override
  void initState() {
    super.initState();
    _observer = Application.of(context).widget.onLabel.observeForever(onLocaleChanged);
  }

  void onLocaleChanged(LabelInterface label) => Dispatcher.postFrame.post(() => setState(() {}));

  @override
  void dispose() {
    _observer.dispose();
    super.dispose();
  }
}
